from pydantic import BaseModel, create_model, Field
from typing import Union, List, Dict, Type, Optional
import logging

logger = logging.getLogger(__name__)


class TypeMapping:
    STRING_TO_TYPE: Dict[str, Type] = {
        "str": str,
        "int": int,
        "float": float,
        "bool": bool,
        "list[str]": List[str],
        "list[int]": List[int],
        "list[float]": List[float],
        "list[bool]": List[bool],
    }

    @classmethod
    def string_to_type(cls, type_string: str) -> Type:
        """Convert a string representation to its corresponding Python type."""
        return cls.STRING_TO_TYPE.get(type_string)


class Slot(BaseModel):
    name: str
    type: str = Field(default="str")
    value: Union[str, int, float, bool, List[str], None] = Field(default=None)
    enum: Optional[list[Union[str, int, float, bool, None]]] = Field(default=[])
    description: str = Field(default="")
    prompt: str = Field(default="")
    required: bool = Field(default=False)
    verified: bool = Field(default=False)


class SlotInput(BaseModel):
    name: str
    value: Union[str, int, float, bool, List[str], None]
    enum: Optional[list[Union[str, int, float, bool, None]]]
    description: str


class SlotInputList(BaseModel):
    slot_input_list: list[SlotInput]


class Verification(BaseModel):
    thought: str
    verification_needed: bool


# format slots for slotfilling input and output
def structured_input_output(slots: list[Slot]) -> tuple[SlotInputList, Type]:
    input_slots = [SlotInput(
        name=slot.name,
        value=slot.value,
        enum=slot.enum,
        description=slot.description
    ) for slot in slots]

    output_format = create_model(
        "DynamicSlotOutputs",
        **{slot.name: Optional[TypeMapping.string_to_type(slot.type)] for slot in slots}
    )
    return SlotInputList(slot_input_list=input_slots), output_format


# format slots after slotfilling
def format_slotfilling_output(slots: list[Slot], response) -> list[Slot]:
    logger.info(f"filled_slots: {response}")
    filled_slots = response.model_dump()
    for slot in slots:
        slot.value = filled_slots[slot.name]
    return slots



import json
import shopify
import logging
import inspect

# general GraphQL navigation utilities
from arklex.env.tools.shopify.utils_nav import *
from arklex.env.tools.shopify.utils import authorify_admin
from arklex.env.tools.shopify.utils_slots import ShopifyCancelOrderSlots, ShopifyOutputs

from arklex.env.tools.tools import register_tool
from arklex.exceptions import ToolExecutionError
from arklex.env.tools.shopify._exception_prompt import ShopifyExceptionPrompt
logger = logging.getLogger(__name__)

description = "Cancel order by order id."
slots = ShopifyCancelOrderSlots.get_all_slots()
outputs = [
    ShopifyOutputs.CANECEL_REQUEST_DETAILS,
]


@register_tool(description, slots, outputs)
def cancel_order(cancel_order_id: str, **kwargs) -> str:
    func_name = inspect.currentframe().f_code.co_name
    auth = authorify_admin(kwargs)
    
    try:
        with shopify.Session.temp(**auth):
            response = shopify.GraphQL().execute(f"""
            mutation orderCancel {{
            orderCancel(
                orderId: "{cancel_order_id}",
                reason: CUSTOMER,
                notifyCustomer: true,
                restock: true,
                refund: true
            ) {{
                userErrors {{
                    field
                    message
                }}
            }}
            }}
            """)
            response = json.loads(response)["data"]
            if not response.get("orderCancel", {}).get("userErrors"):
                return "The order is successfully cancelled. " + json.dumps(response)
            else:
                raise ToolExecutionError(func_name, json.dumps(response["orderCancel"]["userErrors"]))
    
    except Exception as e:
        raise ToolExecutionError(func_name, ShopifyExceptionPrompt.ORDER_CANCEL_ERROR_PROMPT)


import os
import logging
from typing import Dict
import json
from http import HTTPStatus
import argparse
import uvicorn

from openai import OpenAI
from fastapi import FastAPI, Response

from arklex.utils.utils import init_logger
from arklex.env.env import Env
from arklex.orchestrator.orchestrator import AgentOrg
from arklex.utils.model_config import MODEL


logger = logging.getLogger(__name__)
app = FastAPI()


def get_api_bot_response(args, history, user_text, parameters, env):
    data = {"text": user_text, 'chat_history': history, 'parameters': parameters}
    orchestrator = AgentOrg(config=os.path.join(args.input_dir, "taskgraph.json"), env=env)
    result = orchestrator.get_response(data)

    return result['answer'], result['parameters']


@app.post("/eval/chat")
def predict(data: Dict):
    history = data['history']
    params = data['parameters']
    workers = data['workers']
    tools = data['tools']
    user_text = history[-1]['content']

    env = Env(
        tools = tools,
        workers = workers,
        slotsfillapi = ""
    )
    answer, params = get_api_bot_response(args, history[:-1], user_text, params, env)
    return {"answer": answer, "parameters": params}


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Start FastAPI with custom config.")
    parser.add_argument('--input-dir', type=str, default="./examples/test")
    parser.add_argument('--model', type=str, default=MODEL["model_type_or_path"])
    parser.add_argument('--port', type=int, default=8000, help="Port to run the FastAPI app")
    parser.add_argument('--log-level', type=str, default="WARNING", choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    
    args = parser.parse_args()
    os.environ["DATA_DIR"] = args.input_dir
    MODEL["model_type_or_path"] = args.model

    log_level = getattr(logging, args.log_level.upper(), logging.WARNING)
    logger = init_logger(log_level=log_level, filename=os.path.join(os.path.dirname(__file__), "logs", "arklex.log"))

    #run server
    uvicorn.run(app, host="0.0.0.0", port=args.port)


"""
This module is currently inactive.

It is reserved for future use and may contain experimental or planned features.

Status:
    - Not in use (as of 2025-02-20)
    - Intended for future feature expansion

Module Name: chat_server

This file contains the code for setting up a chat server that can accept connections from chat clients and broadcast messages to all connected clients.
"""

import asyncio
import json
import sys

import argparse

class ChatServer:
    ''' Chat server class '''
    
    # dict of all current users
    ALL_USERS = {}
    SERVER_USER = 'Server'
    
    def __init__(self, host_address, host_port):
        self.host_address = host_address
        self.host_port = host_port
        
    # write a message to a stream writer
    async def write_message(self, writer, msg_bytes):
        # write message to this user
        writer.write(msg_bytes)
        # wait for the buffer to empty
        await writer.drain()
    
    # send a message to all connected users
    async def broadcast_message(self, name, message: str=''):
        # report locally
        
        print(f'{name}: {message.strip()}')
        sys.stdout.flush()
        msg_bytes = json.dumps({'name': name, 'message': message}).encode()
        # enumerate all users and broadcast the message
        
        # create a task for each write to client
        tasks = [asyncio.create_task(self.write_message(w, msg_bytes)) for _,(_,w) in self.ALL_USERS.items()]
        # wait for all writes to complete
        if tasks: 
            _ = await asyncio.wait(tasks)
    
    # connect a user
    async def connect_user(self, reader: asyncio.StreamWriter, writer: asyncio.StreamWriter):
        # ask the user for their name
        data = await reader.read(1024)  # Read raw bytes
        # print(data)
        # convert name to string
        name = json.loads(data.decode())['name']
        # store the user details
        self.ALL_USERS[name] = (reader, writer)
        # announce the user
        await self.broadcast_message(self.SERVER_USER, f'{name} has connected\n')
        # welcome message
        await self.write_message(writer, json.dumps({'name': self.SERVER_USER, 'message':  f'Welcome {name}. Send QUIT to disconnect.'}).encode())
        return name
    
    # disconnect a user
    async def disconnect_user(self, name: str, writer: asyncio.StreamWriter):
        # close the user's connection
        writer.close()
        await writer.wait_closed()
        # remove from the dict of all users
        del self.ALL_USERS[name]
        # broadcast the user has left
        await self.broadcast_message(self.SERVER_USER, f'{name} has disconnected\n')
    
    # handle a chat client
    async def handle_chat_client(self, reader: asyncio.StreamWriter, writer: asyncio.StreamWriter):
        print('Client connecting...')
        # connect the user
        name = await self.connect_user(reader, writer)
        try:
            # read messages from the user
            while True:
                # read data
                data = await reader.read(1024)  # Read raw bytes
                decoded_data = data.decode()
                sys.stdout.flush()
                for data in decoded_data.split('{'):
                    if not data:
                        continue
                    
                    # convert to string
                    data_json = json.loads('{' + data)
                    name, line = data_json['name'], data_json['message'].strip()
                    # check for exit
                    if line == 'QUIT':
                        break
                    # broadcast message
                    await self.broadcast_message(name, line)
                else:
                    continue
                break
        finally:
            # disconnect the user
            await self.disconnect_user(name, writer)
    
    # chat server
    async def main(self):
        # define the local host
        # create the server
        server = await asyncio.start_server(self.handle_chat_client, self.host_address, self.host_port)
        # run the server
        async with server:
            # report message
            print('Chat Server Running\nWaiting for chat clients...')
            # accept connections
            await server.serve_forever()

if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('--server-address', type=str, default='127.0.0.1')
    parser.add_argument('--server-port', type=int, default=8888)
    args = parser.parse_args()
    
    server = ChatServer(args.server_address, args.server_port)
    # start the event loop
    asyncio.run(server.main())


from typing import Any, Dict
import json
from http import HTTPStatus
import argparse
import uvicorn

import shopify
from fastapi import FastAPI, Response

from arklex.env.tools.shopify.utils import authorify_admin

USER_NOT_FOUND_ERROR = "error: No user found"
PRODUCTS_NOT_FOUND = "error: No products found"

app = FastAPI()

def get_users(kwargs: Dict[str, Any]) -> str:
    auth = authorify_admin(kwargs)
    if auth["error"]:
        return auth["error"]
    
    try:
        with shopify.Session.temp(**auth["value"]):
            response = shopify.GraphQL().execute(r"""
                {
                    customers(first: 50) {
                        nodes {
                        id
                        firstName
                        lastName
                        email
                        phone
                        createdAt
                        updatedAt
                        numberOfOrders
                        orders(first: 5) {
                            edges {
                                node {
                                id
                                }
                            }
                        }
                        amountSpent {
                            amount
                            currencyCode
                        }
                        lastOrder {
                            id
                            name
                        }
                        addresses {
                            id
                            firstName
                            lastName
                            company
                            address1
                            address2
                            city
                            province
                            country
                            zip
                            phone
                            name
                            provinceCode
                            countryCodeV2
                        }
                        }
                    }
                }
            """)
            data = json.loads(response)['data']['customers']['nodes']
            return data

    except Exception as e:
        print(e)
        return USER_NOT_FOUND_ERROR


def get_products(kwargs) -> str:
    auth = authorify_admin(kwargs)
    if auth["error"]:
        return auth["error"]

    try:
        with shopify.Session.temp(**auth["value"]):
            response = shopify.GraphQL().execute(f"""
                {{
                    products (first: 20) {{
                        nodes {{
                            id
                            title
                            description
                            totalInventory
                            onlineStoreUrl
                            options {{
                                name
                                values
                            }}
                            category {{
                                name
                            }}
                            variants (first: 3) {{
                                nodes {{
                                    displayName
                                    id
                                    price
                                    inventoryQuantity
                                }}
                            }}
                        }}
                        pageInfo {{
                            endCursor
                            hasNextPage
                            hasPreviousPage
                            startCursor
                        }}
                    }}
                }}
            """)
            data = json.loads(response)['data']['products']['nodes']
            response_list = []
            for product in data:
                response_text = ""
                response_text += f"Product ID: {product.get('id', 'None')}\n"
                response_text += f"Title: {product.get('title', 'None')}\n"
                response_text += f"Description: {product.get('description', 'None')}\n"
                response_text += f"Total Inventory: {product.get('totalInventory', 'None')}\n"
                response_text += f"Options: {product.get('options', 'None')}\n"
                response_text += "The following are several variants of the product:\n"
                for variant in product.get('variants', {}).get('nodes', []):
                    response_text += f"Variant name: {variant.get('displayName', 'None')}, Variant ID: {variant.get('id', 'None')}, Price: {variant.get('price', 'None')}, Inventory Quantity: {variant.get('inventoryQuantity', 'None')}\n"
                response_list.append({"id": product.get('id', 'None'), "attribute": response_text})        
            return response_list
    except Exception as e:
        return PRODUCTS_NOT_FOUND
    
@app.get("/users")
def get_users_route():
    users = []
    response = get_users(kwargs)

    for user in response:
        # attribute = f"Your name is {user['firstName']} {user['lastName']} and your email is {user['email']}. Your phone number is {user['phone']}. You registered at {user['createdAt']}. The last time you entered our store was {user['updatedAt']}. You have {user['numberOfOrders']} orders. You have spent {user['amountSpent']['amount']} {user['amountSpent']['currencyCode']} at our store."
        attribute = user
        single_user = {"input": user["id"], "attribute": attribute}
        users.append(single_user)
    return users

@app.get("/products")
def get_products_route():
    products = []
    response = get_products(kwargs)
    for product in response:
        print("============product============")
        print(product)
        # attribute = f"The product is {product['title']}. The description is {product['description']}. The product is in the {product['category']} category."
        single_product = {"input": product["id"], "attribute": product["attribute"]}
        products.append(single_product)
    return products



if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--port", type=int, default=8001)
    args = parser.parse_args()
    
    kwargs = {
        "shop_url": "<your_shop_url>",
        "api_version": "2024-10",
        "admin_token": "<your_admin_token>"
    }
    # print(get_users(kwargs))
    # print(get_products(kwargs))

    #run server
    uvicorn.run(app, host="0.0.0.0", port=args.port)


import os
import json
import argparse

from arklex.evaluation.simulate_first_pass_convos import simulate_conversations
# from arklex.evaluation.simulate_action_convos import simulate_action_conversations
from arklex.evaluation.extract_conversation_info import extract_task_completion_metrics
from arklex.evaluation.simulate_second_pass_convos import get_labeled_convos
# from arklex.evaluation.analyze import analyze_action_accuracy_metrics
from arklex.utils.model_config import MODEL


def evaluate(config):
    task = config['task']
    model_api = config['model_api']
    model_params = config['model_params']
    synthetic_data_params = config['synthetic_data_params']

    bot_goal = config.get('builder_objective', None)
    bot_goal = None if bot_goal == "" else bot_goal
    
    if task == 'first_pass':
        # first pass
        first_pass_data, goals = simulate_conversations(model_api, model_params, synthetic_data_params, config)
        goal_metrics = extract_task_completion_metrics(first_pass_data, bot_goal)
        data = first_pass_data
    elif task == 'action_pass':
        # action pass
        action_pass_data, goals, labels_list = simulate_action_conversations(model_api, model_params, synthetic_data_params, config)
        goal_metrics = extract_task_completion_metrics(action_pass_data, bot_goal)
        goal_metrics['action_pass'] = analyze_action_accuracy_metrics(action_pass_data, labels_list)
        data = action_pass_data
    # second pass
    if task == 'all':
        labeled_convos = get_labeled_convos(first_pass_data, model_api, synthetic_data_params, model_params, config)
    else:
        labeled_convos = []
    return data, labeled_convos, goal_metrics, goals


if __name__ == "__main__":

    parser = argparse.ArgumentParser()
    parser.add_argument('--model_api', type=str)
    parser.add_argument('--model_params', type=dict, default={})
    parser.add_argument('--num_convos', type=int, default=5)
    parser.add_argument('--num_goals', type=int, default=5)
    parser.add_argument('--max_turns', type=int, default=5)
    parser.add_argument('--documents_dir', type=str)
    parser.add_argument('--config', type=str)
    parser.add_argument('--output_dir', type=str)
    parser.add_argument('--model', type=str, default=MODEL["model_type_or_path"])
    parser.add_argument('--customer_type', type=str, default=None, choices=['b2b', 'b2c'])
    parser.add_argument('--task', type=str, default='first_pass', choices=['first_pass', "action_pass", 'all'])
    parser.add_argument('--user_attributes', type=str, default='arklex/evaluation/user_attributes.json')
    parser.add_argument('--custom_profile', action='store_true')
    parser.add_argument('--system_inputs', action='store_true')
    parser.add_argument('--data_file', type=str, default=None)
    args = parser.parse_args()

    MODEL["model_type_or_path"] = args.model

    assert args.model_api is not None, "Model api must be provided"
    assert args.config is not None, "Config file must be provided"
    assert args.documents_dir is not None, "Documents directory must be provided"
    assert args.user_attributes is not None, "User attribute file must be provided"
    if not args.output_dir:
        args.output_dir = os.path.join(args.documents_dir, 'eval')

    if not os.path.exists(args.output_dir):
        os.makedirs(args.output_dir, exist_ok=True)

    config = json.load(open(args.config))
    user_attributes = json.load(open(args.user_attributes))
    # if args.testset:
    #     testset = json.load(open(args.testset))
    # else:
    #     testset = {}
    config['model_api'] = args.model_api
    config['documents_dir'] = args.documents_dir
    config['output_dir'] = args.output_dir
    config['model_params'] = args.model_params
    config['synthetic_data_params'] = {'num_convos': args.num_convos, 'num_goals': args.num_goals, 
                                       'max_turns': args.max_turns, 'customer_type': args.customer_type, 
                                       'data_file': args.data_file}
    config['task'] = args.task
    config['user_attributes'] = user_attributes
    config['custom_profile'] = args.custom_profile
    config['system_inputs'] = args.system_inputs
    first_pass_data, final_convos, goal_metrics, goals = evaluate(config)

    with open(os.path.join(args.output_dir, 'goals.json'), 'w') as f:
        json.dump(goals, f, indent=4)

    with open(os.path.join(args.output_dir, 'simulate_data.json'), 'w') as f:
        json.dump(first_pass_data, f, indent=4)

    with open(os.path.join(args.output_dir, 'labeled_data.json'), 'w') as f:
        json.dump(final_convos, f, indent=4)
    
    with open(os.path.join(args.output_dir, 'goal_completion.json'), 'w') as f:
        json.dump(goal_metrics, f, indent=4)


from datetime import datetime, timezone
import inspect
import hubspot
from hubspot.crm.objects.emails import PublicObjectSearchRequest, ApiException
from hubspot.crm.objects.communications.models import SimplePublicObjectInputForCreate
from hubspot.crm.associations.v4 import AssociationSpec

from arklex.env.tools.tools import register_tool, logger
from arklex.env.tools.hubspot.utils import authenticate_hubspot
from arklex.exceptions import ToolExecutionError
from arklex.env.tools.hubspot._exception_prompt import HubspotExceptionPrompt
description = "Find the contacts record by email. If the record is found, the lastmodifieddate of the contact will be updated. If the correspodning record is not found, the function will return an error message."


slots = [
    {
        "name": "email",
        "type": "str",
        "description": "The email of the user, such as 'something@example.com'.",
        "prompt": "Thanks for your interest in our products! Could you please provide your email or phone number?",
        "required": True,
    },
    {
        "name": "chat",
        "type": "str",
        "description": "This occurs when user communicates with the chatbot",
        "prompt": "",
        "required": True,
    }
]
outputs = [
    {
        "name": "contact_information",
        "type": "str",
        "description": "The basic contact information for the existing customer (e.g. id, first_name, last_name, etc.)",
    }
]


@register_tool(description, slots, outputs)
def find_contact_by_email(email: str, chat: str, **kwargs) -> str:
    func_name = inspect.currentframe().f_code.co_name
    access_token = authenticate_hubspot(kwargs)

    api_client = hubspot.Client.create(access_token=access_token)
    public_object_search_request = PublicObjectSearchRequest(
        filter_groups=[
            {
                "filters": [
                    {
                        "propertyName": "email",
                        "operator": "EQ",
                        "value": email
                    }
                ]
            }
        ]
    )

    try:
        contact_search_response = api_client.crm.contacts.search_api.do_search(
            public_object_search_request=public_object_search_request)
        logger.info("Found contact by email: {}".format(email))
        contact_search_response = contact_search_response.to_dict()
        if contact_search_response['total'] == 1:
            contact_id = contact_search_response['results'][0]['id']
            communication_data = SimplePublicObjectInputForCreate(
                properties = {
                    "hs_communication_channel_type": "CUSTOM_CHANNEL_CONVERSATION",
                    "hs_communication_body": chat,
                    "hs_communication_logged_from": "CRM",
                    "hs_timestamp": datetime.now(timezone.utc).isoformat(),
                }
            )
            contact_info_properties = {
                'contact_id': contact_id,
                'contact_email': email,
                'contact_first_name': contact_search_response['results'][0]['properties'].get('firstname'),
                'contact_last_name': contact_search_response['results'][0]['properties'].get('lastname')
            }
            try:
                communication_creation_response = api_client.crm.objects.communications.basic_api.create(communication_data)
                communication_creation_response = communication_creation_response.to_dict()
                communication_id = communication_creation_response['id']
                association_spec = [
                    AssociationSpec(
                        association_category="HUBSPOT_DEFINED",
                        association_type_id=82
                    )
                ]
                try:
                    association_creation_response = api_client.crm.associations.v4.basic_api.create(
                        object_type="contact",
                        object_id=contact_id,
                        to_object_type="communication",
                        to_object_id=communication_id,
                        association_spec=association_spec
                    )
                except ApiException as e:
                    logger.info("Exception when calling AssociationV4: %s\n" % e)
            except ApiException as e:
                logger.info("Exception when calling basic_api: %s\n" % e)
            return str(contact_info_properties)
        else:
            raise ToolExecutionError(func_name, HubspotExceptionPrompt.USER_NOT_FOUND_PROMPT)
    except ApiException as e:
        logger.info("Exception when calling search_api: %s\n" % e)
        raise ToolExecutionError(func_name, HubspotExceptionPrompt.USER_NOT_FOUND_PROMPT)






import os
import logging
import json
import uuid
import inspect
import traceback

from arklex.utils.graph_state import MessageState, StatusEnum
from arklex.utils.slot import Slot
from arklex.orchestrator.NLU.nlu import SlotFilling
from arklex.utils.utils import format_chat_history
from arklex.exceptions import ToolExecutionError, AuthenticationError

logger = logging.getLogger(__name__)

    
def register_tool(desc, slots=[], outputs=[], isResponse=False):
    current_file_dir = os.path.dirname(__file__)
    def inner(func):
        file_path = inspect.getfile(func)
        relative_path = os.path.relpath(file_path, current_file_dir)
        # reformat the relative path to replace / and \\ with -, and remove .py, because the function calling in openai only allow the function name match the patter the pattern '^[a-zA-Z0-9_-]+$'
        # different file paths format in Windows and linux systems
        relative_path = relative_path.replace("/", "-").replace("\\", "-").replace(".py", "")
        key = f"{relative_path}-{func.__name__}"
        tool = lambda : Tool(func, key, desc, slots, outputs, isResponse)
        return tool
    return inner

class Tool:
    def __init__(self, func, name, description, slots, outputs, isResponse):
        self.func = func
        self.name = name
        self.description = description
        self.output = outputs
        self.slotfillapi: SlotFilling = None
        self.info = self.get_info(slots)
        self.slots = [Slot.model_validate(slot) for slot in slots]
        self.isResponse = isResponse

    def get_info(self, slots):
        self.properties = {}
        for slot in slots:
            self.properties[slot["name"]] = {k: v for k, v in slot.items() if k in ["type", "description", "prompt", "items"]}
        required = [slot["name"] for slot in slots if slot.get("required", False)]
        return {
            "type": "function",
            "function": {
                "name": self.name,
                "description": self.description,
                "parameters": {
                    "type": "object",
                    "properties": self.properties,
                    "required": required,
                },
            },
        }
        
    def init_slotfilling(self, slotfillapi: SlotFilling):
        self.slotfillapi = slotfillapi

    def _init_slots(self, state: MessageState):
        default_slots = state.slots.get("default_slots", [])
        logger.info(f'Default slots are: {default_slots}')
        if not default_slots:
            return
        response = {}
        for default_slot in default_slots:
            response[default_slot.name] = default_slot.value
            for slot in self.slots:
                if slot.name == default_slot.name and default_slot.value:
                    slot.value = default_slot.value
                    slot.verified = True
        state.function_calling_trajectory.append({
            "role": "tool",
            "tool_call_id": str(uuid.uuid4()),
            "name": "default_slots",
            "content": json.dumps(response)
        })
        
        logger.info(f'Slots after initialization are: {self.slots}')
        
    def _execute(self, state: MessageState, **fixed_args):
        # if this tool has been called before, then load the previous slots status
        if state.slots.get(self.name):
            self.slots = state.slots[self.name]
        else:
            state.slots[self.name] = self.slots
        # init slot values saved in default slots
        self._init_slots(state)
        # do slotfilling
        chat_history_str = format_chat_history(state.function_calling_trajectory)
        slots : list[Slot] = self.slotfillapi.execute(self.slots, chat_history_str)
        logger.info(f'{slots=}')
        if not all([slot.value and slot.verified for slot in slots if slot.required]):
            for slot in slots:
                # if there is extracted slots values but haven't been verified
                if slot.value and not slot.verified:
                    # check whether it verified or not
                    verification_needed, thought = self.slotfillapi.verify_needed(slot, chat_history_str)
                    if verification_needed:
                        response = slot.prompt + "The reason is: " + thought
                        break
                    else:
                        slot.verified = True
                # if there is no extracted slots values, then should prompt the user to fill the slot
                if not slot.value:
                    response = slot.prompt
                    break
            
            state.status = StatusEnum.INCOMPLETE

        # if slot.value is not empty for all slots, and all the slots has been verified, then execute the function
        tool_success = False
        if all([slot.value and slot.verified for slot in slots if slot.required]):
            logger.info("all slots filled")
            kwargs = {slot.name: slot.value for slot in slots}
            combined_kwargs = {**kwargs, **fixed_args}
            try:
                response = self.func(**combined_kwargs)
                tool_success = True
            except ToolExecutionError as tee:
                logger.error(traceback.format_exc())
                response = tee.extra_message
            except AuthenticationError as ae:
                logger.error(traceback.format_exc())
                response = str(ae)
            except Exception as e:
                logger.error(traceback.format_exc())
                response = str(e)
            logger.info(f"Tool {self.name} response: {response}")
            call_id = str(uuid.uuid4())
            state.function_calling_trajectory.append({
                'content': None, 
                'role': 'assistant', 
                'tool_calls': [
                    {
                        'function': {
                            'arguments': json.dumps(kwargs), 
                            'name': self.name
                        }, 
                        'id': call_id, 
                        'type': 'function'
                    }
                ], 
                'function_call': None
            })
            state.function_calling_trajectory.append({
                "role": "tool",
                "tool_call_id": call_id,
                "name": self.name,
                "content": response
            })
            state.trajectory[-1][-1].input = slots
            state.trajectory[-1][-1].output = response
            state.status = StatusEnum.COMPLETE if tool_success else StatusEnum.INCOMPLETE

        if self.isResponse and tool_success:
            logger.info("Tool output is stored in response instead of message flow")
            state.response = response
        else:
            state.message_flow = state.message_flow + f"Context from {self.name} tool execution: {response}\n"
        state.slots[self.name] = slots
        return state

    def execute(self, state: MessageState, **fixed_args):
        state = self._execute(state, **fixed_args)
        return state
    
    def __str__(self):
        return f"{self.__class__.__name__}"

    def __repr__(self):
        return f"{self.__class__.__name__}"
    


import json
import random
import requests
import copy
from arklex.evaluation.get_documents import load_docs
from arklex.evaluation.chatgpt_utils import (chatgpt_chatbot, query_chatbot, filter_convo, adjust_goal,
                                               flip_hist, generate_goals, format_chat_history_str, flip_hist_content_only)
from arklex.env.env import Env
from arklex.orchestrator.NLU.nlu import SlotFilling
from arklex.env.tools.tools import Tool

ATTR_TO_PROFILE = "Convert the following list user attributes in to a text description of a customer profile for the following company:\n{company_summary}\nThe user attributes are here:\n{user_attr}"
ADAPT_GOAL = "Assume you are planning to speak to a chatbot with the following goal in mind:\n{goal}\nUsing the company information below, re-write this goal into one that is more specific to the company. The new goal should mention specific products (if relevant) or other details about the company. Here is a summary of the company:\n{company_summary}\nHere is a page from the company website:\n{company_doc}"
ADD_ATTRIBUTES = "Your job is to add attributes to a customer profile. Here is an example of an existing profile with the categories on the left and the attributes on the right:\n{user_profile}\nSuggest three attributes for the following category:\n{category}\nThese attributes should be specific values that are relevant to the category and apply to potential customers of the company. You should return a comma separated list of attributes without any descriptions of the attributes. Generated the attributes based on a summary of the company and the company webpage and what kind of customers the compnay is likely targeting. Here is the summary fo the company:\n{company_summary}\nHere is the webpage:\n{company_doc}"


def build_profile(synthetic_data_params, config) -> tuple[list[str], list[str], list[dict], list[dict], list[dict]]:
    labels_list = []
    if not config['custom_profile']: # Use predefined profiles (from user_attributes.json)
        documents = load_docs(config['documents_dir'], config, synthetic_data_params['num_goals'] * 2)
        attributes_list = filter_attributes(config)
        attributes_list = augment_attributes(attributes_list, config, documents)
        attributes_list = select_attributes(attributes_list, synthetic_data_params)
        if config['system_inputs']:
            system_attributes_list = select_system_attributes(config, synthetic_data_params)
        else:
            system_attributes_list = [{}] * synthetic_data_params['num_convos']
        attributes_list_with_goals = adapt_goals(attributes_list, config, documents)
        profiles, goals, system_inputs = convert_attributes_to_profiles(attributes_list_with_goals, system_attributes_list, config)
        return profiles, goals, attributes_list, system_inputs, labels_list

    else: # Use custom profiles (from database)
        user_profiles, system_attributes = get_custom_profiles(config)
        predefined_attributes = filter_attributes(config)
        system_attributes_list = []
        attributes_list = []
        labels_list = []
        for i in range(synthetic_data_params['num_convos']):
            system_attribute = {}
            user_profile = {}
            binding_index = {}
            for key, value in system_attributes.items():
                full_key = f"system_attributes.{key}"
                if "bind_to" in value:
                    random_index = random.choice(range(len(system_attributes[key])))
                    system_attribute[key] = system_attributes[key][random_index]
                    binding_index[value["bind_to"]] = random_index
                else:
                    random_index = random.choice(range(len(system_attributes[key])))
                    system_attribute[key] = system_attributes[key][random_index]
        
            for key, value in user_profiles.items():
                full_key = f"user_profiles.{key}"
                if "bind_to" in value and full_key in binding_index:
                    user_profile[key] = user_profiles[key][binding_index[full_key]]
                if "bind_to" not in value:
                    random_index = random.choice(range(len(user_profiles[key])))
                    user_profile[key] = user_profiles[key][random_index]
            # based on the user's profile, select the attribute
            attribute = pick_attribute(user_profile, predefined_attributes)
            # get the proposed tool from the goal and the corresponding input as label
            label, valid = get_label(attribute, config)
            
            labels_list.append(label)
            attributes_list.append(attribute)
            system_attributes_list.append(system_attribute)
        
        profiles, goals, system_inputs = convert_attributes_to_profiles(attributes_list, system_attributes_list, config)
        
    return profiles, goals, attributes_list, system_inputs, labels_list


def pick_attribute(user_profile, predefined_attributes):
    """
    Pick the attribute from the predefined attributes based on the user's profile to avoid the attribute confliction
    """
    attributes = {}
    for key, value in predefined_attributes.items():
        PICK_ATTRIBUTE_PROMPT = """Given the user's profile, please pick or modify the attribute value from the choice list of the given attribute category {category}. If all the given choices are not align with the user's profile, then generate a new attribute value that is most likely to be used by the user. Please directly return the attribute value without any description.
        User's profile: {user_profile}
        {category}: {choices}
        Attribute value:
        """
        response = chatgpt_chatbot([{'role': 'system', 'content': PICK_ATTRIBUTE_PROMPT.format(user_profile=user_profile, category=key, choices="\n".join(value))}])
        attributes[key] = response
        
    return attributes


def get_custom_profiles(config) -> tuple[dict, dict]:

    if "system_attributes" in config["user_attributes"] and "user_profiles" in config["user_attributes"]:
        # First, get system attributes with bindings
        system_attributes = {}
        bindings = {}  # Track bindings between fields
        
        # Process system_attributes and their bindings
        for key, value in config["user_attributes"]["system_attributes"].items():
            full_key = f"system_attributes.{key}"
            if isinstance(value, dict):
                api_url = value.get("api")
                response = requests.get(api_url).json()
                system_attributes[key] = response
                # Track bindings if they exist
                if "bind_to" in value:
                    bindings[full_key] = response
                    bindings[value["bind_to"]] = response
            else:
                system_attributes[key] = value
    
        user_profiles = {}
        # Process user_profiles and their bindings
        for key, value in config["user_attributes"]["user_profiles"].items():
            if isinstance(value, dict):
                if "bind_to" in value and value["bind_to"] in bindings:
                    user_profiles[key] = bindings[value["bind_to"]]
                else:
                    api_url = value.get("api")
                    response = requests.get(api_url).json()
                    user_profiles[key] = response
            else:
                user_profiles[key] = value
    
    return user_profiles, system_attributes


def get_label(attribute, config):
    """
    Get the appropriate tool used by the Agent to achieve the user's goal
    """
    valid = True   #dummy variable
    GET_TOOL_PROMPT = """Given the list of tools that an AI assistant can use, and the user's goal, return the tool that is most likely to be used to achieve the goal. Only return the tool id.
    Tools: {tools}
    User's goal: {goal}
    Tool_id:
    """
    env = Env(
        tools = config["tools"],
        workers = config["workers"]
    )
    tool_list = []
    for tool in config["tools"]:
        tool_id = tool["id"]
        tool: Tool = env.tools[tool_id]["execute"]()
        slots = tool.slots
        tool_description = tool.description
        tool_input =[s.model_dump() for s in slots]
        tool_output = tool.output
        tool_list.append({
            "tool_id": tool_id,
            "tool_description": tool_description,
            "tool_input": tool_input,
            "tool_output": tool_output
        })
    tool_list.append({
        "tool_id": "0",
        "tool_description": "There are no tools appropriate for the goal.",
        "tool_input": [],
        "tool_output": "There are no tools appropriate for the goal."
    })

    label = [
        {
            "tool_id": "0",
            "tool_name": "No tool",
            "slots": {},
        }
    ]
    attempt = 0
    while attempt < 3:
        try:
            response = chatgpt_chatbot(
                [{'role': 'system', 'content': GET_TOOL_PROMPT.format(tools="\n".join(f"tool_id: {tool['tool_id']}\ntool_description: {tool['tool_description']}\ntool_input: {tool['tool_input']}\ntool_output: {tool['tool_output']}" for tool in tool_list), goal=attribute["goal"])}]
            )
            pred_tool_id = response
            if pred_tool_id == "0":
                break
            selected_tool = env.tools[pred_tool_id]["execute"]()
            slots = selected_tool.slots
            pred_slots = SlotFilling(url="").execute(slots, str(attribute), type="user_simulator")
            pred_slots_dict = {slot.name: slot.value for slot in pred_slots}
            label = [
                {
                    "tool_id": pred_tool_id,
                    "tool_name": selected_tool.name,
                        "slots": pred_slots_dict,
                }
            ]
            break
        except Exception as e:
            attempt += 1


    return label, valid

def filter_attributes(config) -> dict:
    filtered_attributes = {}
    for key in config['user_attributes'].keys():
        if key == 'generic' or key == config['synthetic_data_params']['customer_type']:
            for subkey in config['user_attributes'][key].keys():
                filtered_attributes[subkey] = config['user_attributes'][key][subkey]
    return filtered_attributes

def select_system_attributes(config, synthetic_data_params) -> list[dict[str, dict]]:
    system_attributes = []
    for subkey, subvalue in config["user_attributes"]["system_attributes"].items():
        if isinstance(subvalue, dict):
            api_url = subvalue.get("api")
            response = requests.get(api_url).json()
            config["user_attributes"]["system_attributes"][subkey] = response

    for i in range(synthetic_data_params['num_convos']):
        system_attribute = {}
        for subkey, subvalue in config["user_attributes"]["system_attributes"].items():
            if isinstance(subvalue, list) and isinstance(subvalue[0], dict):
                system_attribute[subkey] = random.choice(subvalue)
            else:
                raise ValueError("System attributes should be a list of dictionaries")
        system_attributes.append(copy.deepcopy(system_attribute))
    return system_attributes


def select_attributes(user_attributes, synthetic_data_params):
    user_list = []
    for i in range(synthetic_data_params['num_convos']):
        attributes = {}
        for key, value in user_attributes.items():
            attributes[key] = random.choice(value)
        user_list.append(attributes.copy())
    return user_list

def augment_attributes(attributes_list, config, documents):
    # add attribute values using docs 
    new_attrs = generate_attributes(attributes_list, config, documents)
    return new_attrs

def adapt_goals(attributes_list, config, documents):
    attributes_list_with_goals = []
    for item in attributes_list:
        new_goal = adapt_goal(item['goal'], config, documents)
        new_item = {}
        for key in item.keys():
            if key == 'goal':
                new_item['goal'] = new_goal
            else:
                new_item[key] = item[key]
        attributes_list_with_goals.append(new_item)
    return attributes_list_with_goals

def adapt_goal(goal, config, documents):
    new_goal = chatgpt_chatbot([{'role': 'user', 'content': ADAPT_GOAL.format(goal=goal, company_summary=config['intro'], company_doc=random.choice(documents))}])
    return new_goal

def generate_attributes(attributes, config, documents):
    text_attribute = ''
    for key, value in attributes.items():
        if len(value['values']) == 0:
            continue
        text_attribute += f"{key}: {value['values']}\n"

    new_attrs = {}
    for category in attributes.keys():
        if not attributes[category]['generate_values']:
            new_attrs[category] = attributes[category]['values']
        else:
            attrs = chatgpt_chatbot([{'role': 'user', 'content': ADD_ATTRIBUTES.format(user_profile=text_attribute, category=category, company_summary=config['intro'], company_doc=random.choice(documents))}])
            new_attrs[category] = attrs.split(', ')
    return new_attrs

def attributes_to_text(attribute_list):
    text_attributes = []
    for item in attribute_list:
        text_attribute = ''
        for key, value in item.items():
            text_attribute += f"{key}: {value}\n"
        text_attributes.append(text_attribute[:-1])
    return text_attributes

def convert_attributes_to_profiles(attributes_list, system_attributes, config, mode="full"):
    profile_list = []
    system_inputs = []
    for sys_attr, attr_list in zip(system_attributes, attributes_list):
        system_input = {}
        for key, value in sys_attr.items():
            if mode == "full":
                attr_list[key] = value["attribute"]
            system_input[key] = value["input"]
        system_inputs.append(system_input)

    text_attributes = attributes_to_text(attributes_list)
    for i, attribute in enumerate(text_attributes):
        profile = chatgpt_chatbot([{'role': 'user', 'content': ATTR_TO_PROFILE.format(company_summary=config['intro'], user_attr=attribute)}])
        profile_list.append({"profile": profile, "goal": attributes_list[i]["goal"]})
    
    profiles = [item['profile'] for item in profile_list]
    goals = [item['goal'] for item in profile_list]
    return profiles, goals, system_inputs


def build_labelled_profile(synthetic_data_params, config):
    env = Env(
        tools = config["tools"],
        workers = config["workers"]
    )
    
    user_profiles, system_attributes = get_custom_profiles(config, synthetic_data_params)
    labels_list = []
    attributes_list_with_goals = []
    system_attributes_list = []
    for i in range(synthetic_data_params['num_convos']):
        system_attribute = {}
        user_profile = {}
        binding_index = {}
        for key, value in system_attributes.items():
            full_key = f"system_attributes.{key}"
            if "bind_to" in value:
                random_index = random.choice(range(len(system_attributes[key])))
                system_attribute[key] = system_attributes[key][random_index]
                binding_index[value["bind_to"]] = random_index
            else:
                random_index = random.choice(range(len(system_attributes[key])))
                system_attribute[key] = system_attributes[key][random_index]
    
        for key, value in user_profiles.items():
            full_key = f"user_profiles.{key}"
            if "bind_to" in value and full_key in binding_index:
                user_profile[key] = user_profiles[key][binding_index[full_key]]
            if "bind_to" not in value:
                random_index = random.choice(range(len(user_profiles[key])))
                user_profile[key] = user_profiles[key][random_index]
        # option1: 
        # randomly choose any one of tools from the whole tool list for each conversation
        # select the direct slot value from the user's profile
        while True:
            # slot filling
            tool_id = random.choice(config["tools"])["id"]
            tool: Tool = env.tools[tool_id]["execute"]()
            slots = tool.slots
            pred_slots = SlotFilling(url="").execute(slots, str(user_profile), metadata={}, type="user_simulator")
            # if all slots are filled, means that the user's profile has enough information to complete the goal
            # if not, try again
            if all(slot.value for slot in pred_slots):
                break
        pred_slots_dict = {slot.name: slot.value for slot in pred_slots}
        label = [
            {
                "tool_id": tool_id,
                "tool_name": tool.name,
                    "slots": pred_slots_dict,
            }
        ]
        labels_list.append(label)

        tool_description = tool.description
        tool_input =[s.model_dump() for s in pred_slots]
        tool_output = tool.output
        goal_generation_prompt = f"""Given a tool that an AI assistant can use, imagine what kind of user query or request would naturally require this tool to fulfill it.

Tool Description: {tool_description}
Tool Input: {tool_input}
Tool Output: {tool_output}

Think about:
1. What problem or need would a user have that this tool could solve?
2. What would the user naturally ask for, without knowing this tool exists?
3. What real-world goal would lead to using this tool?

Generate a natural high-level user goal that would require this tool to fulfill it. No example or detailed content is needed.
Write it as a first-person statement starting with "I want" or "I need". A user can only do the action on themselves. Cannot manipulate the tool to do the action on other people.
The goal should within 15 words.


User's goal:"""
        goal = chatgpt_chatbot(
            [{'role': 'system', 'content': goal_generation_prompt}]
        )
        attributes_list_with_goals.append({"goal": goal, **pred_slots_dict})
        print("+++++++++++++++++++++++++++")
        print("tool_description: ", tool_description)
        print("goal: ", goal)
        print("+++++++++++++++++++++++++++")

        system_attributes_list.append(system_attribute)
    
    
    profiles, goals, system_inputs = convert_attributes_to_profiles(attributes_list_with_goals, system_attributes_list, config, mode="separate")
    
    
    return profiles, goals, attributes_list_with_goals, system_inputs, labels_list
    


import ast
import json
from datetime import datetime, timedelta
import pytz
import inspect

import hubspot
import parsedatetime
from hubspot.crm.objects.meetings import ApiException

from arklex.env.tools.tools import register_tool, logger
from arklex.env.tools.hubspot.utils import authenticate_hubspot
from arklex.exceptions import ToolExecutionError
from arklex.env.tools.hubspot._exception_prompt import HubspotExceptionPrompt


description = "Schedule a meeting for the existing customer with the specific representative."


slots = [
    {
        "name": "cus_fname",
        "type": "str",
        "description": "The first name of the customer contact.",
        "prompt": "",
        "required": True,
    },
    {
        "name": "cus_lname",
        "type": "str",
        "description": "The last name of the customer contact.",
        "prompt": "",
        "required": True,
    },
    {
        "name": "cus_email",
        "type": "str",
        "description": "The email of the customer contact.",
        "prompt": "",
        "required": True,
    },
    {
        "name": "meeting_date",
        "type": "str",
        "description": "The exact date the customer want to take meeting with the representative. e.g. tomorrow, today, Next Monday, May 1st. If you are not sure about the input, ask the user to give you confirmation.",
        "prompt": "Could you please give me the date of the meeting?",
        "required": True,
    },
    {
        "name": "meeting_start_time",
        "type": "str",
        "description": "The exact start time the customer want to take meeting with the representative. e.g. 1pm, 1:00 PM. If you are not sure about the input, ask the user to give you confirmation.",
        "prompt": "Could you please give me the start time of the meeting? Typically, the representative will hold the meeting from 9:00 am to 4:45 pm.",
        "required": True,
    },
    {
        "name": "duration",
        "type": "int",
        "enum": [15, 30, 60],
        "description": "The exact duration of the meeting. Please ask the user to input. DO NOT AUTOMATICALLY GIVE THE SLOT ANY VALUE.",
        "prompt": "Could you please give me the duration of the meeting (e.g. 15, 30, 60 mins)?",
        "required": True,
    },
    {
        "name": "slug",
        "type": "str",
        "description": "The corresponding slug for the meeting link, which is extracted from check_available tool",
        "prompt": "",
        "required": True,
    },
    {
        "name": "bt_slots_ux",
        "type": "str",
        "description": "The busy time slots (unix form) of the representative. This is extracted from check_available tool. This is a list of dict.",
        "prompt": "",
        "required": True,
    },
    {
        "name": "time_zone",
        "type": "str",
        "enum": ["America/New_York", "America/Los_Angeles", "Asia/Tokyo", "Europe/London"],
        "description": "The timezone of the user. For example, 'America/New_York'.",
        "prompt": "Could you please provide your timezone or where are you now?",
        "required": True
    }
]
outputs = [
    {
        "name": "meeting_confirmation_info",
        "type": "string",
        "description": "The detailed information about the meeting to let the customer confirm",
    }
]


@register_tool(description, slots, outputs)
def create_meeting(cus_fname: str, cus_lname: str, cus_email: str, meeting_date: str,
                   meeting_start_time: str, duration: int,
                   slug: str, bt_slots_ux: str, time_zone: str, **kwargs) -> str:
    func_name = inspect.currentframe().f_code.co_name
    access_token = authenticate_hubspot(kwargs)

    meeting_date = parse_natural_date(meeting_date, timezone=time_zone, date_input=True)
    meeting_start_time = parse_natural_date(meeting_start_time, meeting_date, timezone=time_zone)
    meeting_start_time = int(meeting_start_time.timestamp() * 1000)


    duration = int(duration)
    duration = int(timedelta(minutes=duration).total_seconds() * 1000)

    meeting_end_time = meeting_start_time + duration

    bt_slots_ux = json.loads(bt_slots_ux)
    for time_slot in bt_slots_ux:
        if meeting_start_time >= time_slot['start'] and meeting_start_time < time_slot['end']:
            raise ToolExecutionError(func_name, HubspotExceptionPrompt.MEETING_UNAVAILABLE_PROMPT)
        elif meeting_end_time >= time_slot['start'] and meeting_end_time <= time_slot['end']:
            raise ToolExecutionError(func_name, HubspotExceptionPrompt.MEETING_UNAVAILABLE_PROMPT)

    api_client = hubspot.Client.create(access_token=access_token)

    try:
        create_meeting_response = api_client.api_request(
            {
                "path": "/scheduler/v3/meetings/meeting-links/book",
                "method": "POST",
                "body": {
                    "slug": slug,
                    "duration": duration,
                    "startTime": meeting_start_time,
                    "email": cus_email,
                    "firstName": cus_fname,
                    "lastName": cus_lname,
                    "timezone": time_zone,
                    "locale": "en-us",
                },
                "qs": {
                    'timezone': time_zone
                }
            }

        )
        create_meeting_response = create_meeting_response.json()
        return json.dumps(create_meeting_response)
    except ApiException as e:
        logger.info("Exception when scheduling a meeting: %s\n" % e)
        raise ToolExecutionError(func_name, HubspotExceptionPrompt.MEETING_UNAVAILABLE_PROMPT)


def parse_natural_date(date_str, base_date=None, timezone=None, date_input=False):
    cal = parsedatetime.Calendar()
    time_struct, _ = cal.parse(date_str, base_date)
    if date_input:
        parsed_dt = datetime(*time_struct[:3])
    else:
        parsed_dt = datetime(*time_struct[:6])

    if base_date and (parsed_dt.date() != base_date.date()):
        parsed_dt = datetime.combine(base_date.date(), parsed_dt.time())

    if timezone:
        local_timezone = pytz.timezone(timezone)
        parsed_dt = local_timezone.localize(parsed_dt)
        parsed_dt = parsed_dt.astimezone(pytz.utc)
    return parsed_dt





import os
import json
import argparse
import time
import logging
from dotenv import load_dotenv
from pprint import pprint

from arklex.utils.utils import init_logger
from arklex.orchestrator.orchestrator import AgentOrg
from arklex.utils.model_config import MODEL
from arklex.utils.model_provider_config import LLM_PROVIDERS
from arklex.env.env import Env

load_dotenv()

def pprint_with_color(data, color_code="\033[34m"):  # Default to blue
    print(color_code, end="")  # Set the color
    pprint(data)
    print("\033[0m", end="")  


def get_api_bot_response(args, history, user_text, parameters, env):
    data = {"text": user_text, 'chat_history': history, 'parameters': parameters}
    orchestrator = AgentOrg(config=os.path.join(args.input_dir, "taskgraph.json"), env=env)
    result = orchestrator.get_response(data)

    return result['answer'], result['parameters'], result['human_in_the_loop']


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('--input-dir', type=str, default="./examples/test")
    parser.add_argument('--model', type=str, default=MODEL["model_type_or_path"])
    parser.add_argument( '--llm-provider',type=str,default=MODEL["llm_provider"],choices=LLM_PROVIDERS)
    parser.add_argument('--log-level', type=str, default="WARNING", choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"])
    args = parser.parse_args()
    os.environ["DATA_DIR"] = args.input_dir
    MODEL["model_type_or_path"] = args.model
    MODEL["llm_provider"] = args.llm_provider
    log_level = getattr(logging, args.log_level.upper(), logging.WARNING)
    logger = init_logger(log_level=log_level, filename=os.path.join(os.path.dirname(__file__), "logs", "arklex.log"))

    # Initialize env
    config = json.load(open(os.path.join(args.input_dir, "taskgraph.json")))
    env = Env(
        tools = config.get("tools", []),
        workers = config.get("workers", []),
        slotsfillapi = config["slotfillapi"]
    )
        
    history = []
    params = {}
    user_prefix = "user"
    worker_prefix = "assistant"
    for node in config['nodes']:
        if node[1].get("type", "") == 'start':
            start_message = node[1]['attribute']["value"]
            break
    history.append({"role": worker_prefix, "content": start_message})
    pprint_with_color(f"Bot: {start_message}")
    
    while True:
        user_text = input("You: ")
        if user_text.lower() == "quit":
            break
        start_time = time.time()
        output, params, hitl = get_api_bot_response(args, history, user_text, params, env)
        history.append({"role": user_prefix, "content": user_text})
        history.append({"role": worker_prefix, "content": output})
        print(f"getAPIBotResponse Time: {time.time() - start_time}")
        pprint_with_color(f"Bot: {output}")



import json
import time
from typing import Any, Dict
import logging
import uuid
import os
from typing import List, Dict, Any, Tuple
import ast
import copy
import janus
from dotenv import load_dotenv

from langchain_core.runnables import RunnableLambda

from arklex.env.env import Env
from arklex.orchestrator.task_graph import TaskGraph
from arklex.env.tools.utils import ToolGenerator
from arklex.types import StreamType
from arklex.utils.graph_state import (ConvoMessage, NodeInfo, OrchestratorMessage,
                                      MessageState, PathNode,StatusEnum,
                                      BotConfig, Params, ResourceRecord,
                                      OrchestratorResp)
from arklex.utils.utils import format_chat_history


load_dotenv()
logger = logging.getLogger(__name__)

INFO_WORKERS = ["planner", "MessageWorker", "RagMsgWorker"]

class AgentOrg:
    def __init__(self, config, env: Env, **kwargs):
        if isinstance(config, dict):
            self.product_kwargs = config
        else:
            self.product_kwargs = json.load(open(config))
        self.user_prefix = "user"
        self.worker_prefix = "assistant"
        self.environment_prefix = "tool"
        self.__eos_token = "\n"
        self.task_graph = TaskGraph("taskgraph", self.product_kwargs)
        self.env = env

    
    def init_params(self, inputs) -> Tuple[str, str, Params]:
        text = inputs["text"]
        chat_history = inputs["chat_history"]
        input_params = inputs["parameters"]

        # Create base params with defaults
        params = Params()
        
        # Update with any provided values
        if input_params:
            params = Params.model_validate(input_params)
        
        # Update specific fields
        chat_history_copy = copy.deepcopy(chat_history)
        chat_history_copy.append({"role": self.user_prefix, "content": text})
        chat_history_str = format_chat_history(chat_history_copy)        
        # Update turn_id and function_calling_trajectory
        params.metadata.turn_id += 1
        if not params.memory.function_calling_trajectory:
            params.memory.function_calling_trajectory = copy.deepcopy(chat_history_copy)
        else:
            params.memory.function_calling_trajectory.extend(chat_history_copy[-2:])
        
        return text, chat_history_str, params

    def check_skip_node(self, node_info: NodeInfo, params: Params):
        return False
        if not node_info.can_skipped:
            return False
        cur_node_id = params.taskgraph.curr_node
        if cur_node_id in params.taskgraph.node_limit:
            if params.taskgraph.node_limit[cur_node_id] <= 0:
                return True
        return False

    def post_process_node(self, node_info: NodeInfo, params: Params, update_info:dict={}):
        '''
        update_info is a dict of
            skipped = Optional[bool]
        '''
        curr_node = params.taskgraph.curr_node
        node = PathNode(
            node_id = curr_node,
            is_skipped = update_info.get("is_skipped", False),
            in_flow_stack=node_info.add_flow_stack,
            nested_graph_node_value = None,
            nested_graph_leaf_jump = None,
        )
        
        params.taskgraph.path.append(node)

        if curr_node in params.taskgraph.node_limit:
            params.taskgraph.node_limit[curr_node] -= 1
        return params
    
    def handl_direct_node(self, node_info: NodeInfo, params: Params):
        # Direct response
        node_attribute = node_info.attributes
        if node_attribute.get("value", "").strip() and node_attribute.get("direct"):
            params = self.post_process_node(node_info, params)
            return_response = OrchestratorResp(
                answer=node_attribute["value"],
                parameters=params.model_dump()
            )
            # TODO: multiple choice list
            # if node_attribute.get("type", "") == "multiple-choice" and node_attribute.get("choice_list", []):
            #     return_response["choice_list"] = node_attribute["choice_list"]
            return True, return_response, params
        return False, None, params
    
    def perform_node(self, node_info: NodeInfo, params: Params, text: str, chat_history_str: str, same_turn: bool, stream_type: StreamType, message_queue: janus.SyncQueue):
        # Tool/Worker
        user_message = ConvoMessage(history=chat_history_str, message=text)
        orchestrator_message = OrchestratorMessage(message=node_info.attributes["value"], attribute=node_info.attributes)
        sys_instruct = "You are a " + self.product_kwargs["role"] + ". " + \
                            self.product_kwargs["user_objective"] + \
                            self.product_kwargs["builder_objective"] + \
                            self.product_kwargs["intro"] + \
                            self.product_kwargs.get("opt_instruct", "")
        bot_config = BotConfig(
            bot_id=self.product_kwargs.get("bot_id", "default"),
            version=self.product_kwargs.get("version", "default"),
            language=self.product_kwargs.get("language", "EN"),
            bot_type=self.product_kwargs.get("bot_type", "presalebot"),
        )
    
        # Create initial resource record with common info and output from trajectory
        resource_record = ResourceRecord(
            info={
                "id": node_info.resource_id,
                "name": node_info.resource_name,
                "attribute": node_info.attributes,
                "node_id": params.taskgraph.curr_node
            }
        )

        # If this is a new turn, create a new list
        if not same_turn:
            params.memory.trajectory.append([])
        
        # Add resource record to current turn's list
        params.memory.trajectory[-1].append(resource_record)

        message_state = MessageState(
            sys_instruct=sys_instruct, 
            bot_config=bot_config,
            user_message=user_message, 
            orchestrator_message=orchestrator_message, 
            function_calling_trajectory=params.memory.function_calling_trajectory, 
            trajectory=params.memory.trajectory,
            message_flow=params.memory.trajectory[-1][-1].output if params.memory.trajectory[-1] else "", 
            slots=params.taskgraph.dialog_states,
            metadata=params.metadata,
            is_stream=True if stream_type is not None else False,
            message_queue=message_queue
        )
        
        response_state, params = self.env.step(node_info.resource_id, message_state, params)
        params.memory.trajectory = response_state.trajectory
        return response_state, params
    
    def _get_response(self, 
                     inputs: dict, 
                     stream_type: StreamType = None, 
                     message_queue: janus.SyncQueue = None) -> OrchestratorResp:
        text, chat_history_str, params = self.init_params(inputs)
        ##### TaskGraph Chain
        same_turn = False
        taskgraph_inputs = {
            "text": text,
            "chat_history_str": chat_history_str,
            "parameters": params,
            "allow_global_intent_switch": True,
        }
        taskgraph_chain = RunnableLambda(self.task_graph.get_node) | RunnableLambda(self.task_graph.postprocess_node)

        # TODO: when planner is re-implemented, execute/break the loop based on whether the planner should be used (bot config).
        msg_counter = 0
        
        n_node_performed = 0
        max_n_node_performed = 5
        while n_node_performed < max_n_node_performed:
            taskgraph_start_time = time.time()
            node_info, params = taskgraph_chain.invoke(taskgraph_inputs)
            taskgraph_inputs["allow_global_intent_switch"] = False
            params.metadata.timing.taskgraph = time.time() - taskgraph_start_time
            # Check if current node can be skipped
            can_skip = self.check_skip_node(node_info, params)
            if can_skip:
                params = self.post_process_node(node_info, params, {"is_skipped": True})
                continue
            logger.info(f"The current node info is : {node_info}")
            # Check current node attributes
            if node_info.resource_name in INFO_WORKERS:
                msg_counter += 1
            # handle direct node
            is_direct_node, direct_response, params = self.handl_direct_node(node_info, params)
            if is_direct_node:
                return direct_response
            # perform node
            response_state, params = self.perform_node(node_info,
                                                       params,
                                                       text,
                                                       chat_history_str,
                                                       same_turn,
                                                       stream_type,
                                                       message_queue)
            params = self.post_process_node(node_info, params)
            n_node_performed += 1
            same_turn = True
            # If the current node is not complete, then no need to continue to the next node
            node_status = params.taskgraph.node_status
            cur_node_id = params.taskgraph.curr_node
            status = node_status.get(cur_node_id, StatusEnum.COMPLETE)
            if status == StatusEnum.INCOMPLETE:
                break
            # If the counter of message worker or counter of planner or counter of ragmsg worker == 1, break the loop
            if msg_counter == 1:
                break
            if node_info.is_leaf is True:
                break

        if not response_state.response:
            logger.info("No response, do context generation")
            if not stream_type:
                response_state = ToolGenerator.context_generate(response_state)
            else:
                response_state = ToolGenerator.stream_context_generate(response_state)
        
        # TODO: Need to reformat the RAG response from trajectory
        # params["memory"]["tool_response"] = {}
        return OrchestratorResp(
            answer=response_state.response,
            parameters=params.model_dump(),
            human_in_the_loop=params.metadata.hitl,
        )
    
    def get_response(self, 
                     inputs: dict, 
                     stream_type: StreamType = None, 
                     message_queue: janus.SyncQueue = None) -> Dict[str, Any]:
        orchestrator_response = self._get_response(inputs, stream_type, message_queue)
        return orchestrator_response.model_dump()



from abc import ABC, abstractmethod
from arklex.utils.graph_state import MessageState, StatusEnum
import logging
import traceback

logger = logging.getLogger(__name__)

def register_worker(cls):
    """Decorator to register a worker."""
    cls.name = cls.__name__  # Automatically set name to the class name
    return cls


class BaseWorker(ABC):
    
    description = None

    def __str__(self):
        return f"{self.__class__.__name__}"

    def __repr__(self):
        return f"{self.__class__.__name__}"
    
    @abstractmethod
    def _execute(self, msg_state: MessageState):
        pass

    def execute(self, msg_state: MessageState):
        try:
            response_return = self._execute(msg_state)
            response_state = MessageState.model_validate(response_return)
            response_state.trajectory[-1][-1].output = response_state.response if response_state.response else response_state.message_flow
            if response_state.status == StatusEnum.INCOMPLETE:
                response_state.status = StatusEnum.COMPLETE
        except Exception as e:
            logger.error(traceback.format_exc())
            response_state.status = StatusEnum.INCOMPLETE
        return response_state



from typing import Any, Optional, List, Dict
from pydantic import BaseModel, Field
from enum import Enum
import uuid
from arklex.utils.slot import Slot

### Bot-related classes
class BotConfig(BaseModel):
    bot_id: str
    version: str
    language: str
    bot_type: str

### Message-related classes

class ConvoMessage(BaseModel):
    history: str # it could be the whole original message or the summarization of the previous conversation from memory module
    message: str


class OrchestratorMessage(BaseModel):
    message: str
    attribute: dict


### Task status-related classes

class StatusEnum(str, Enum):
    COMPLETE = "complete"
    INCOMPLETE = "incomplete"
    STAY = "stay"

class Timing(BaseModel):
    taskgraph: Optional[float] = None

class ResourceRecord(BaseModel):
    info: Dict
    input: List = Field(default_factory=list)
    output: str = Field(default="")
    steps: List = Field(default_factory=list)

class Metadata(BaseModel):
    # TODO: May need to initialize the metadata(i.e. chat_id, turn_id) based on the conversation database
    chat_id: str = Field(default_factory=lambda: str(uuid.uuid4()))
    turn_id: int = 0
    hitl: Optional[str] = Field(default=None)
    timing: Timing = Field(default_factory=Timing)

class MessageState(BaseModel):
    # system configuration
    sys_instruct: str
    # bot configuration 
    bot_config: BotConfig
    # input message
    user_message: ConvoMessage
    orchestrator_message: OrchestratorMessage
    # action trajectory
    function_calling_trajectory: List[Dict[str, Any]]
    trajectory: List[List[ResourceRecord]]
    # message flow between different nodes
    message_flow: str = Field(description="message flow between different nodes")
    # final response
    response: str = Field(default="")
    # task-related params
    status: StatusEnum = Field(default=StatusEnum.INCOMPLETE)
    slots: Dict[str, List[Slot]] = Field(description="record the dialogue states of each action")
    metadata: Metadata
    # stream
    is_stream: bool
    message_queue: Any = Field(exclude=True)


class PathNode(BaseModel):
    node_id: str
    is_skipped: bool = False
    in_flow_stack: bool = False
    nested_graph_node_value: Optional[str] = None
    nested_graph_leaf_jump: Optional[str] = None


class Taskgraph(BaseModel):
    # Need add global intent
    dialog_states: Dict[str, List[Slot]] = Field(default_factory=dict)
    path: List[PathNode] = Field(default_factory=list)
    curr_node: str = Field(default="")
    intent: str = Field(default="")
    node_limit: Dict[str, int] = Field(default_factory=dict)
    nlu_records: List = Field(default_factory=list)
    node_status: Dict[str, StatusEnum] = Field(default_factory=dict)
    available_global_intents: List = Field(default_factory=list)


class Memory(BaseModel):
    trajectory: List[List[ResourceRecord]] = Field(default_factory=list)
    function_calling_trajectory: List[Dict[str, Any]] = Field(default_factory=list)
    
class Params(BaseModel):
    metadata: Metadata = Field(default_factory=Metadata)
    taskgraph: Taskgraph = Field(default_factory=Taskgraph)
    memory: Memory = Field(default_factory=Memory)

class NodeInfo(BaseModel):
    resource_id: str = Field(default="")
    resource_name: str = Field(default="")
    can_skipped: bool = Field(default=False)
    is_leaf: bool = Field(default=False)
    attributes: Dict[str, Any] = Field(default_factory=dict)
    add_flow_stack: Optional[bool] = Field(default=False)

class OrchestratorResp(BaseModel):
    answer: str = Field(default="")
    parameters: Dict[str, Any] = Field(default_factory=dict)
    human_in_the_loop: Optional[str] = Field(default=None)


__all__ = [
    "AuthenticationError",
    "ToolExecutionError",
    "ExceptionPrompt",
]


class AuthenticationError(Exception):
    """
    Exception raised when authentication fails.
    """
    def __init__(self, message: str):
        self.message = f"Authentication failed: {message}"
        super().__init__(self.message)


class UserFacingError(Exception):
    """
    Exception raised to guide the user to update their query.
    """
    def __init__(self, message: str, extra_message: str):
        super().__init__(message)
        # Store the additional message in a custom attribute, which will be used to guide the user to update their query.
        self.extra_message = extra_message


class ToolExecutionError(UserFacingError):
    """
    Exception raised when a tool execution fails.
    """
    def __init__(self, message: str, extra_message: str):
        self.message = f"Tool {message} execution failed"
        super().__init__(self.message, extra_message)


class ExceptionPrompt:
    """
    Base class for tool-specific exception prompts.
    
    This class serves as a parent class for tool collections (like Shopify, HubSpot)
    to define their own exception prompts as class attributes.
    
    Example:
        class ShopifyExceptionPrompt(ExceptionPrompt):
            ORDER_NOT_FOUND = "Order could not be found."
            PRODUCT_NOT_AVAILABLE = "Product is not available."
    
    Each tool collection should create their own _exception_prompt.py file
    that inherits from this base class.
    """

    # Common exception prompts shared across tool collections
    pass



from arklex.exceptions import AuthenticationError

HUBSPOT_AUTH_ERROR = "Missing some or all required hubspot authentication parameters: access_token. Please set up 'fixed_args' in the config file. For example, {'name': <unique name of the tool>, 'fixed_args': {'token': <shopify_access_token>, 'shop_url': <shopify_shop_url>, 'api_version': <Shopify API version>}}"


def authenticate_hubspot(kwargs):
    access_token = kwargs.get('access_token')
    if not access_token:
        raise AuthenticationError(HUBSPOT_AUTH_ERROR)

    return access_token




import sys
from pathlib import Path
sys.path.append(str(Path(__file__).resolve().parents[3]))

import logging
import string

from fastapi import FastAPI, Response

from arklex.utils.slot import Verification, SlotInputList, structured_input_output, format_slotfilling_output, Slot
from dotenv import load_dotenv
load_dotenv()

from arklex.utils.model_config import MODEL
from arklex.utils.model_provider_config import PROVIDER_MAP
from langchain_openai import ChatOpenAI
from pydantic_ai import Agent


logger = logging.getLogger(__name__)


class NLUModelAPI ():
    def __init__(self):
        self.user_prefix = "user"
        self.assistant_prefix = "assistant"

    def get_response(self, sys_prompt, model, response_format="text", note="intent detection"):
        logger.info(f"Prompt for {note}: {sys_prompt}")
        dialog_history = [{"role": "system", "content": sys_prompt}]
        kwargs = {'model': MODEL["model_type_or_path"], 'temperature': 0.7}
        
        if MODEL['llm_provider'] != 'anthropic': kwargs['n'] = 1
        llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(**kwargs)

        if MODEL['llm_provider'] == 'openai':
            llm = llm.bind(response_format={"type": "json_object"} if response_format == "json" else {"type": "text"})
            res = llm.invoke(dialog_history)
        else:
            messages = [("user", f"{dialog_history[0]['content']} Only choose the option letter, no explanation.")]
            res = llm.invoke(messages)
        
        return res.content

    def format_input(self, intents, chat_history_str) -> str:
        """Format input text before feeding it to the model."""
        intents_choice, definition_str, exemplars_str = "", "", ""
        idx2intents_mapping = {}
        multiple_choice_index = dict(enumerate(string.ascii_lowercase))
        count = 0
        for intent_k, intent_v in intents.items():
            if len(intent_v) == 1:
                intent_name = intent_k
                idx2intents_mapping[multiple_choice_index[count]] = intent_name
                definition = intent_v[0].get("attribute", {}).get("definition", "")
                sample_utterances = intent_v[0].get("attribute", {}).get("sample_utterances", [])

                if definition:
                    definition_str += (
                        f"{multiple_choice_index[count]}) {intent_name}: {definition}\n"
                    )
                if sample_utterances:
                    exemplars = "\n".join(sample_utterances)
                    exemplars_str += (
                        f"{multiple_choice_index[count]}) {intent_name}: \n{exemplars}\n"
                    )
                intents_choice += f"{multiple_choice_index[count]}) {intent_name}\n"

                count += 1

            else:
                for idx, intent in enumerate(intent_v):
                    intent_name = f'{intent_k}__<{idx}>'
                    idx2intents_mapping[multiple_choice_index[count]] = intent_name
                    definition = intent.get("attribute", {}).get("definition", "")
                    sample_utterances = intent.get("attribute", {}).get("sample_utterances", [])

                    if definition:
                        definition_str += (
                            f"{multiple_choice_index[count]}) {intent_name}: {definition}\n"
                        )
                    if sample_utterances:
                        exemplars = "\n".join(sample_utterances)
                        exemplars_str += (
                            f"{multiple_choice_index[count]}) {intent_name}: \n{exemplars}\n"
                        )
                    intents_choice += f"{multiple_choice_index[count]}) {intent_name}\n"

                    count += 1
        # Base prompt without conditional sections
        system_prompt_nlu = """According to the conversation, decide what is the user's intent in the last turn? \n"""
        
        # Conditionally add definitions if available
        if definition_str.strip():
            system_prompt_nlu += """Here are the definitions for each intent:\n{definition}\n"""
            
        # Conditionally add exemplars if available
        if exemplars_str.strip():
            system_prompt_nlu += """Here are some sample utterances from user that indicate each intent:\n{exemplars}\n"""
            
        # Add the rest of the prompt
        system_prompt_nlu += """Conversation:\n{formatted_chat}\n\nOnly choose from the following options.\n{intents_choice}\n\nAnswer:"""
        
        system_prompt = system_prompt_nlu.format(
            definition=definition_str,
            exemplars=exemplars_str,
            intents_choice=intents_choice,
            formatted_chat=chat_history_str,
        )
        return system_prompt, idx2intents_mapping

    def predict(
        self,
        text,
        intents,
        chat_history_str,
        model
    ) -> str:

        system_prompt, idx2intents_mapping = self.format_input(
            intents, chat_history_str
        )
        response = self.get_response(
            system_prompt, model, note="intent detection"
        )
        logger.info(f"postprocessed intent response: {response}")
        try:
            pred_intent_idx = response.split(")")[0]
            pred_intent = idx2intents_mapping[pred_intent_idx]
        except:
            pred_intent = response.strip().lower()
        logger.info(f"postprocessed intent response: {pred_intent}")
        return pred_intent


class SlotFillModelAPI():
    def __init__(self):
        self.user_prefix = "user"
        self.assistant_prefix = "assistant"
    
    # System prompt for slot filling
    def format_input(self, slots: SlotInputList, input, type: str = "chat") -> str:
        """Format input text before feeding it to the model."""
        if type == "chat":
            system_prompt = f"Given the conversation and definition of each dialog state, update the value of following dialogue states.\nDialogue Statues:\n{slots}\nConversation:\n{input}\n\n"
        elif type == "user_simulator":
            system_prompt = f"Given a user profile, extract the values for each defined slot type. Only extract values that are explicitly mentioned in the profile. If a value is not found, leave it empty.\n\nSlot definitions:\n{slots}\n\nUser profile:\n{input}\n\nFor each slot:\n1. Look for an exact match in the profile\n2. Only extract values that are clearly stated\n3. Do not make assumptions or infer values\n4. If a slot has enum values, the extracted value must match one of them exactly\n\nExtract the values:\n"
        return system_prompt

    # get response from model
    def get_response(self, sys_prompt, format, note="slot filling"):
        logger.info(f"Prompt for {note}: {sys_prompt}")
        dialog_history = [{"role": "system", "content": sys_prompt}]
        kwargs = {'model': MODEL["model_type_or_path"], 'temperature': 0.7}
        # set number of chat completions to generate, isn't supported by Anthropic
        if MODEL['llm_provider'] != 'anthropic': kwargs['n'] = 1
        llm = PROVIDER_MAP.get(MODEL['llm_provider'], ChatOpenAI)(**kwargs)
        
        if MODEL['llm_provider'] == 'openai':
            llm = llm.with_structured_output(schema=format)
            response = llm.invoke(dialog_history)
    
        # TODO: fix slotfilling for huggingface
        elif MODEL['llm_provider']=='huggingface':
            # llm = llm.bind_tools([format])
            # chain =  llm | JsonOutputToolsParser()
            # response = chain.invoke(dialog_history)
            raise NotImplementedError("Slotfilling for Huggingface is not implemented")

        elif MODEL['llm_provider'] == 'gemini':
            agent = Agent(f"google-gla:{MODEL['model_type_or_path']}", result_type=format)
            result = agent.run_sync(dialog_history[0]['content'])
            response = result.data

        #for claude 
        else:
            messages = [{"role": "user", "content": dialog_history[0]['content']}]
            llm = llm.bind_tools([format])
            res= llm.invoke(messages)
            response = format(**res.tool_calls[0]['args'])
        return response

    # endpoint for slot filling
    def predict(
        self,
        slots: list[Slot],
        input: str,
        type: str = "chat"
    ):
        input_slots, output_slots = structured_input_output(slots)
        system_prompt = self.format_input(input_slots, input, type)
        response = self.get_response(system_prompt, output_slots, note="slot filling")
        filled_slots = format_slotfilling_output(slots, response)
        logger.info(f"Updated dialogue states: {filled_slots}")
        return filled_slots
    
    # endpoint for slot verification
    def verify(
        self,
        slot: dict,
        chat_history_str,
    ) -> Verification:
        reformat_slot = {key: value for key, value in slot.items() if key in ["name", "type", "value", "enum", "description", "required"]}
        system_prompt = f"Given the conversation, definition and extracted value of each dialog state, decide whether the following dialog states values need further verification from the user. Verification is needed for expressions which may cause confusion. If it is an accurate information extracted, no verification is needed. If there is a list of enum value, which means the value has to be chosen from the enum list. Only Return boolean value: True or False. \nDialogue Statues:\n{reformat_slot}\nConversation:\n{chat_history_str}\n\n"
        response = self.get_response(
            system_prompt, format=Verification, note="slot verification"
        )
        if not response: # no need to verification, we want to make sure it is really confident that we need to ask the question again
            logger.info(f"Failed to verify dialogue states")
            return Verification(verification_needed=False, thought="No need to verify")
        logger.info(f"Verified dialogue states: {response}")
        return response


app = FastAPI()
nlu_api = NLUModelAPI()
slotfilling_api = SlotFillModelAPI()


@app.post("/nlu/predict")
def predict(data: dict, res: Response):
    logger.info(f"Received data: {data}")
    pred_intent = nlu_api.predict(**data)

    logger.info(f"pred_intent: {pred_intent}")
    return {"intent": pred_intent}

@app.post("/slotfill/predict")
def predict(data: dict, res: Response):
    logger.info(f"Received data: {data}")
    results = slotfilling_api.predict(**data)
    logger.info(f"pred_slots: {results}")
    return results

@app.post("/slotfill/verify")
def verify(data: dict, res: Response):
    logger.info(f"Received data: {data}")
    verify_needed = slotfilling_api.verify(**data)

    logger.info(f"verify_needed: {verify_needed}")
    return verify_needed


from arklex.exceptions import ExceptionPrompt


class HubspotExceptionPrompt(ExceptionPrompt):
    """
    HubSpot-specific exception prompts.
    """
    # check_available exception prompt
    MEETING_LINK_UNFOUND_PROMPT = "The representative does not have a meeting link."

    # create_meeting exception prompt
    MEETING_UNAVAILABLE_PROMPT = "The representative is not available during the required period."

    # create_ticket exception prompt
    TICKET_CREATION_ERROR_PROMPT = "Ticket creation failed, please try again later."

    # find_contact_by_email exception prompt
    USER_NOT_FOUND_PROMPT = "User not found (not an existing customer)"

    # find_owner_id_by_contact_id exception prompt
    OWNER_UNFOUND_PROMPT = 'Owner not found (not an existing customer)'


import json
import os
import random
from arklex.evaluation.get_documents import load_docs
from arklex.evaluation.build_user_profiles import build_profile, ATTR_TO_PROFILE
from arklex.evaluation.chatgpt_utils import (chatgpt_chatbot, query_chatbot, filter_convo, adjust_goal,
                                               flip_hist, generate_goals, format_chat_history_str, flip_hist_content_only)
from arklex.env.tools.tools import Tool

# USER_DATA_KEYS = ['goal', 'product_experience_level', 'deal_stage', 'customer_type', 'decision_making_authority', 'persona', 'discovery_type', 'buying_behavior']
USER_DATA_KEYS = ['goal', 'product_experience_level', 'customer_type', 'persona', 'discovery_type', 'buying_behavior']

def get_relevant_vals(attr):
    vals = []
    for key in USER_DATA_KEYS:
        vals.append(attr[key])
    return vals

def count_matches(l1,l2):
    num_matches = 0
    for i in range((len(l1))):
        if l1[i] == l2[i]:
            num_matches += 1
    return num_matches

def join_messages(messages):
    message_str = ""
    for message in messages:
        if message['role'] == 'bot_follow_up':
            continue
        message_str += f"{message['role']}: {message['content']}\n"
    return message_str[:-1]

def create_convo_profile(best_match, attr_vals, summary):
    dict_profile = {}
    for i in range(len(USER_DATA_KEYS)):
        if best_match[i] == 'other':
            continue
        dict_profile[USER_DATA_KEYS[i]] = best_match[i]
    
    text_profile = ''
    for key, value in dict_profile.items():
        text_profile += f"{key}: {value}\n"
    profile = chatgpt_chatbot([{'role': 'user', 'content': ATTR_TO_PROFILE.format(company_summary=summary, user_attr=text_profile[:-1])}])
    return profile

def retrieve_convo(attr_vals, all_profiles, user_convos, summary):
    split_profiles = [p.split(',') for p in all_profiles]
    best_match = None
    max_matches = 0
    for profile in split_profiles:
        num_matches = count_matches(attr_vals, profile)
        if num_matches >= max_matches:
            best_match = profile
            max_matches = num_matches
            num_matches = 0

    convo = random.choice(user_convos[','.join(best_match)])
    convo_messages = join_messages(convo['message'])
    convo_profile = create_convo_profile(best_match, attr_vals, summary)
    return convo_messages, convo_profile

def get_example_convo(attr, synthetic_data_params, summary):
    with open(synthetic_data_params['data_file']) as f:
        user_convos = json.load(f)

    all_profiles = list(user_convos.keys())
    attr_vals = get_relevant_vals(attr)
    convo, matching_profile = retrieve_convo(attr_vals, all_profiles, user_convos, summary)
    return convo, matching_profile

def retrieve_prompts(profile, goal, attr, summary, synthetic_data_params):
    if synthetic_data_params['data_file'] is None:
        instructional_prompt = f'Pretend you are a human interacting with a customer service chatbot for the following company: {summary}\nYou have the following goal when interacting with this chatbot:\n{goal}\nHere is a description of the customer you are pretending to be:\n{profile}\nHave a conversation with the chatbot while trying to achieve your goal as this customer. Make sure the conversation is natural. For example, if the chatbot asks you a question you should answer it.'
        start_text = "Humans write short questions with occasional typos. Here are some examples of what a human customer would type: [how much is it?, Can you send info to my email, yes I need a job, want to check both proposals to rent and buy, How much does it cost a [PRODUCT_HERE], Im interested in [PRODUCT_HERE], hi i would like to rent out [PRODUCT_HERE] but im wondering which countries are available for rental]. Replicate the writing behavior of a human customer and begin the conversation with a question to achieve your goal."
    else:
        example_convo, matching_profile = get_example_convo(attr, synthetic_data_params, summary)
        instructional_prompt = f'Pretend you are a human interacting with a customer service chatbot for the following company: {summary}\nYou have the following goal when interacting with this chatbot:\n{goal}\nHere is a description of the customer you are pretending to be:\n{profile}\nHave a conversation with the chatbot while trying to achieve your goal as this customer. Make sure the conversation is natural. For example, if the chatbot asks you a question you should answer it. Below is an example conversation between a user with a similar profile to yours that you can use a guide. However, keep in mind that the users profile may not be the exact same as yours, so take that into consideration when conducting the conversation. Here is the sample users profile:\n{matching_profile}\nAnd here is the conversation between this user and the chatbot:\n{example_convo}'
        start_text = "Replicate the writing behavior of a human customer and begin the conversation with a question to achieve your goal."
    return instructional_prompt, start_text

def check_goal_completion(goal, convo):
    convo_str = format_chat_history_str(flip_hist_content_only(convo[2:]))
    prompt = f"Here is a conversation between a user and a customer service chatbot assistant:\n{convo_str}\n\nThe user's goal is the following: {goal}\nOutput False if the user needs to learn more information regarding their goal. Output True otherwise. Only onput True or False and nothing else."
    output = chatgpt_chatbot([{'role': 'user', 'content': prompt}])
    return output == "True"

def conversation(model_api, profile, goal, attr, sys_input, summary, model_params, synthetic_data_params, env_config):
    instructional_prompt, start_text = retrieve_prompts(profile, goal, attr, summary, synthetic_data_params)
    history = []
    history.append({'role': 'system','content': instructional_prompt})
    history.append({'role': 'user', 'content': start_text})
    chatbot_history = []
    default_slots = []
    for key, value in sys_input.items():
        if key and value:
            default_slots.append({"name": key, "value": value})
    model_params = {"taskgraph": {"dialog_states": {"default_slots": default_slots}}}
    goal_completetion = False

    for i in range(synthetic_data_params['max_turns']):
        output = chatgpt_chatbot(history) 
        history.append({'role': 'assistant', 'content': output})
        chatbot_history.append({'role': 'assistant', 'content': output})
        curr_node = model_params.get('curr_node', "start") 
        response_data = query_chatbot(model_api, chatbot_history, model_params, env_config)
        answer = response_data["answer"]
        answer = answer.replace('\n', ' ')
        model_params = response_data["parameters"]
        history[-1]['intent'] = model_params["taskgraph"]["intent"]  ## TODO: After add global intent, change to global intent, the current intent if the last intent of each turn
        history[-1]['curr_node'] = model_params["taskgraph"]["curr_node"]
        history[-1]['trajectory'] = model_params["memory"]["trajectory"][-1]

        history.append({'role': 'user', 'content': answer})
        chatbot_history.append({'role': 'user', 'content': answer})
        if i > 2 and check_goal_completion(goal, history.copy()):
            goal_completetion = True
            # history.append({'goal_completetion': True})
            break
    
    # if not history[-1].get('goal_completetion', False):
    #     history.append({'goal_completetion': False})
    return history, goal_completetion

def generate_conversations(model_api, profiles, goals, attributes_list, system_inputs, summary, model_params, synthetic_data_params, env_config):
    convos = []
    for profile, goal, attr, sys_input in zip(profiles, goals, attributes_list, system_inputs):
        convo, goal_completion = conversation(model_api, profile, goal, attr, sys_input, summary, model_params, synthetic_data_params, env_config)
        syn_convo = flip_hist(filter_convo(convo, filter_turns=False))
        convos.append(
            {
                "convo": syn_convo,
                "profile": profile,
                "goal": goal,
                "attributes": attr,
                "goal_completion": goal_completion
            }
        )
    return convos

def simulate_conversations(model_api, model_params, synthetic_data_params, config):
    profiles, goals, attributes_list, system_inputs, labels_list = build_profile(synthetic_data_params, config)

    # save the profiles, goals, attributes_list, system_inputs, labels_list in a json file
    os.makedirs(os.path.join(config['output_dir'], "simulate_data"), exist_ok=True)
    with open(os.path.join(config['output_dir'], "simulate_data", "profiles.json"), "w") as f:
        json.dump(profiles, f, indent=4)
    with open(os.path.join(config['output_dir'], "simulate_data", "goals.json"), "w") as f:
        json.dump(goals, f, indent=4)
    with open(os.path.join(config['output_dir'], "simulate_data", "attributes_list.json"), "w") as f:
        json.dump(attributes_list, f, indent=4)
    with open(os.path.join(config['output_dir'], "simulate_data", "system_inputs.json"), "w") as f:
        json.dump(system_inputs, f, indent=4)
    with open(os.path.join(config['output_dir'], "simulate_data", "labels_list.json"), "w") as f:
        json.dump(labels_list, f, indent=4)

    # with open(os.path.join(config['output_dir'], "simulate_data", "profiles.json"), "r") as f:
    #     profiles = json.load(f)
    # with open(os.path.join(config['output_dir'], "simulate_data", "goals.json"), "r") as f:
    #     goals = json.load(f)
    # with open(os.path.join(config['output_dir'], "simulate_data", "attributes_list.json"), "r") as f:
    #     attributes_list = json.load(f)
    # with open(os.path.join(config['output_dir'], "simulate_data", "system_inputs.json"), "r") as f:
    #     system_inputs = json.load(f)
    # with open(os.path.join(config['output_dir'], "simulate_data", "labels_list.json"), "r") as f:
    #     labels_list = json.load(f)
    summary = config['intro']
    env_config = {
        "workers": config['workers'],
        "tools": config["tools"]
    }
    
    # try:
    conversations = generate_conversations(
        model_api,
        profiles,
        goals,
        attributes_list,
        system_inputs,
        summary,
        model_params,
        synthetic_data_params,
        env_config,
    )
    # except Exception as e:
    #     print("Generate conversations failed")
    #     print("Error: ", e)
    #     conversations = []
    return conversations, goals

if __name__ == "__main__":
    model_api = "http://adaptation.cs.columbia.edu:55231/qa/richtech/v1alpha1"
    synthetic_data_params = {'num_convos': 5, 'num_goals': 3, 'max_turns': 10}
    model_params = {}
    convos  = simulate_conversations(model_api, model_params, synthetic_data_params)
    with open('p1_sample_convos.json', 'w') as f:
        json.dump(convos, f, indent=5)